#include "create_hmac.h"
#include "hmac_testcases.h"

#include <tchar.h>
#include <stdlib.h>
#include <stdio.h>
#include <assert.h>

static int run_single_test (HCRYPTPROV hProv, ALG_ID Algid,
					 		BYTE *data, DWORD data_size,
							BYTE *key,  DWORD key_size,
							const BYTE *hash, DWORD hash_size)
{
	HCRYPTHASH hmac_hash = 0;
	HCRYPTKEY hmac_key = 0;
	BYTE output[1024];
	DWORD output_size = 1024;
	BOOL status;

	status = CreateHMAC (hProv, Algid,
		key, key_size,
		/* flags = */ 0,
		/* phHash = */ &hmac_hash, /* phKey = */ &hmac_key);
	if (!status)
	{
		fprintf (stderr, "CreateHMAC failed with status %#x.\n",
			GetLastError ());
		return (-1);
	}

	status = CryptHashData (hmac_hash,
		data, data_size,
		/* flags = */ 0);
	if (!status)
	{
		fprintf (stderr, "CryptHashData failed with status %#x.\n",
			GetLastError ());
		CryptDestroyHash (hmac_hash);
		CryptDestroyKey (hmac_key);
		return (-1);
	}

	memset (output, 0, sizeof (output));
	status = CryptGetHashParam (hmac_hash, HP_HASHVAL,
		output, &output_size, /* flags = */ 0);
	if (!status)
	{
		fprintf (stderr, "CryptHashData failed with status %#x.\n",
			GetLastError ());
		CryptDestroyHash (hmac_hash);
		CryptDestroyKey (hmac_key);
		return (-1);
	}

	CryptDestroyHash (hmac_hash);
	CryptDestroyKey (hmac_key);

	if (output_size != hash_size)
	{
		fprintf (stderr, "Hash size mismatch: Got %lu, expected %lu.\n",
			output_size, hash_size);
		return (-1);
	}

	if (memcmp (output, hash, hash_size) != 0)
	{
		fprintf (stderr, "Hash sum mismatch.\n");
		return (-1);
	}
	else
	{
		printf ("    Success!\n");
	}

	return (0);
} /* int run_single_test */

static int run_testcase (HCRYPTPROV hProv,
						 const testcase_t *tc)
{
	printf ("  SHA-256:\n");
	run_single_test (hProv, CALG_SHA_256,
		tc->data, tc->data_size,
		tc->key, tc->key_size,
		tc->sha256, sizeof (tc->sha256));
	printf ("\n");

	printf ("  SHA-384:\n");
	run_single_test (hProv, CALG_SHA_384,
		tc->data, tc->data_size,
		tc->key, tc->key_size,
		tc->sha384, sizeof (tc->sha384));
	printf ("\n");

	printf ("  SHA-512:\n");
	run_single_test (hProv, CALG_SHA_512,
		tc->data, tc->data_size,
		tc->key, tc->key_size,
		tc->sha512, sizeof (tc->sha512));
	printf ("\n");

	return (0);
} /* int run_testcase */

int _tmain(int argc, _TCHAR* argv[])
{
	HCRYPTPROV cry_provider;
	BOOL status;

	status = CryptAcquireContext (&cry_provider,
		/* szContainer = */ NULL,
		/* CSP name = */ NULL,
		/* provider type = */ PROV_RSA_AES,
		/* flags = */ CRYPT_VERIFYCONTEXT);
	if (!status)
	{
		fprintf (stderr, "CryptAcquireContext failed: %#x\n", GetLastError ());
		exit (EXIT_FAILURE);
	}

	printf ("Testcase 1\n");
	printf ("----------\n");
	run_testcase (cry_provider, &testcase1);
	printf ("\n");

	printf ("Testcase 2\n");
	printf ("----------\n");
	run_testcase (cry_provider, &testcase2);
	printf ("\n");

	printf ("Testcase 3\n");
	printf ("----------\n");
	run_testcase (cry_provider, &testcase3);
	printf ("\n");

	printf ("Testcase 6\n");
	printf ("----------\n");
	run_testcase (cry_provider, &testcase6);
	printf ("\n");

	CryptReleaseContext (cry_provider, /* flags = */ 0);

	return 0;
}

